import { isAmdEnv } from "../lib/amd-utils";
import { constants, FL_NAMESPACE } from "../constants";
import assetLoader from "@braintree/asset-loader";
import { AxoSupportedPlatforms } from "../types";
import { btModulesLoadConfig } from "../lib/safeLoadBtModule/types";
import { isEsmSupported, loadModule } from "../lib/esm-utils";
import { loadScriptWithExport } from "../lib/script-utils";

jest.mock("@braintree/asset-loader", () => {
  return {
    loadScript: jest.fn().mockResolvedValue(undefined),
    loadStylesheet: jest.fn().mockResolvedValue(undefined),
  };
});

jest.mock("../lib/esm-utils");
const mockIsEsmSupported = isEsmSupported as jest.MockedFn<typeof isAmdEnv>;
const mockLoadModule = loadModule as jest.MockedFn<typeof loadModule>;

jest.mock("../lib/amd-utils");
const mockIsAmdEnv = isAmdEnv as jest.MockedFn<typeof isAmdEnv>;

jest.mock("../lib/script-utils");
const mockLoadScriptWithExport = loadScriptWithExport as jest.MockedFn<
  typeof loadScriptWithExport
>;

import loadAxo from "../loadAxoScript";

describe("loadAxoScript", () => {
  let testContext;
  const version = "3.96";
  const originalWindow = window;
  const originalPerformance = performance;
  const originalGetElementsByTagName = document.getElementsByTagName;
  let mockFastlane;

  beforeEach(() => {
    mockFastlane = jest.fn();

    testContext = {
      mockLoadResponse: {
        fastlane: mockFastlane,
        metadata: {
          localeUrl: "https://www.paypalobjects.com/connect-boba/locales/",
        },
      },
      client: {
        getVersion: () => {
          return version;
        },
      },
    };

    mockLoadModule.mockResolvedValue({ default: mockFastlane });
    mockLoadScriptWithExport.mockResolvedValue(mockFastlane);

    // Mock performance API
    Object.defineProperty(globalThis, "performance", {
      value: {
        mark: jest.fn(),
        measure: jest.fn(),
      },
    });
  });

  afterEach(() => {
    Object.defineProperty(globalThis, "window", {
      value: originalWindow,
    });

    Object.defineProperty(globalThis, "performance", {
      value: originalPerformance,
    });

    Object.defineProperty(document, "getElementsByTagName", {
      value: originalGetElementsByTagName,
    });
  });

  describe("Unhappy path", () => {
    it("rejects promise if platform is invalid", async () => {
      const promise = loadAxo({
        // @ts-expect-error for test purposes
        platform: "Invalid Platform",
        btSdkVersion: version,
        minified: true,
      });
      await expect(promise).rejects.toEqual(
        new Error("unsupported axo platform")
      );
    });
  });

  describe("in a non-AMD environment", () => {
    beforeAll(() => {
      mockIsAmdEnv.mockReturnValue(false);
    });

    it("should load Bt hcf and axo when platform is bt", async () => {
      await loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
        minified: true,
      });

      expect(assetLoader.loadScript).toHaveBeenCalledTimes(1);
      expect(assetLoader.loadScript).toHaveBeenCalledWith(
        expect.objectContaining({
          id: `hcf-${version}`,
          src: `https://js.braintreegateway.com/web/${version}/js/hosted-fields.min.js`,
        })
      );
    });

    it("should load Bt client, hcf and Axo when platform is ppcp", async () => {
      await loadAxo({
        platform: AxoSupportedPlatforms.PPCP,
        btSdkVersion: version,
        minified: true,
      });

      expect(assetLoader.loadScript).toHaveBeenCalledTimes(2);

      expect(assetLoader.loadScript).toHaveBeenCalledWith(
        expect.objectContaining({
          id: `client-${version}`,
          src: `https://js.braintreegateway.com/web/${version}/js/client.min.js`,
        })
      );
      expect(assetLoader.loadScript).toHaveBeenCalledWith(
        expect.objectContaining({
          id: `hcf-${version}`,
          src: `https://js.braintreegateway.com/web/${version}/js/hosted-fields.min.js`,
        })
      );
    });

    describe("when ESM is supported", () => {
      beforeAll(() => {
        mockIsEsmSupported.mockReturnValue(true);
      });

      it("should call loadModule with the minified URL when specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });

      it("should load the unminified URL when specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: false,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.unminified}.js`
        );
      });

      it("should default to the minified script when no environment is specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });

      it("should load the bundle override URL when specified", async () => {
        const bundleId = "2bc4343c92655";
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          metadata: { bundleIdOverride: bundleId },
        });

        await expect(promise).resolves.toStrictEqual({
          fastlane: mockFastlane,
          metadata: {
            localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
          },
        });

        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });
    });

    describe("when ESM is NOT supported", () => {
      beforeAll(() => {
        mockIsEsmSupported.mockReturnValue(false);
      });

      it("should call loadScript with the minified URL when specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
      });
    });

    it("should load the unminified URL when specified", async () => {
      const promise = loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
        minified: false,
      });

      await expect(promise).resolves.toStrictEqual(
        testContext.mockLoadResponse
      );
      expect(mockLoadScriptWithExport).toHaveBeenCalled();
      expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
        expect.objectContaining({
          src: `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.unminified}.js`,
        })
      );
    });

    it("should default to the minified script when no environment is specified", async () => {
      const promise = loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
      });

      await expect(promise).resolves.toStrictEqual(
        testContext.mockLoadResponse
      );
      expect(mockLoadScriptWithExport).toHaveBeenCalled();
      expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
        expect.objectContaining({
          src: `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
        })
      );
    });

    it("should load the bundle override URL when specified", async () => {
      const bundleId = "2bc4343c92655";
      const promise = loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
        metadata: { bundleIdOverride: bundleId },
      });

      await expect(promise).resolves.toStrictEqual({
        fastlane: mockFastlane,
        metadata: {
          localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
        },
      });
      expect(mockLoadScriptWithExport).toHaveBeenCalled();
      expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
        expect.objectContaining({
          src: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
        })
      );
    });

    it("should call loadScript with the minified URL and bundle override URL when specified", async () => {
      const bundleId = "9df9369c92655";
      const promise = loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
        minified: true,
        metadata: { bundleIdOverride: bundleId },
      });

      await expect(promise).resolves.toStrictEqual({
        fastlane: mockFastlane,
        metadata: {
          localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
        },
      });
      expect(mockLoadScriptWithExport).toHaveBeenCalled();
      expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
        expect.objectContaining({
          src: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
        })
      );
    });
  });

  describe("in an AMD environment", () => {
    const originalRequire = require;
    let mockRequire: jest.Mock;

    beforeEach(() => {
      mockIsAmdEnv.mockReturnValue(true);
      mockRequire = jest.fn().mockImplementation((deps, cb) => {
        cb();
      });
      Object.defineProperty(globalThis, "require", {
        value: mockRequire,
        writable: true,
      });
    });

    afterEach(() => {
      Object.defineProperty(globalThis, "require", {
        value: originalRequire,
      });
    });

    it("should load Bt hcf when platform is bt", async () => {
      await loadAxo({
        platform: AxoSupportedPlatforms.BT,
        btSdkVersion: version,
        minified: true,
      });

      expect(mockRequire).toHaveBeenCalledTimes(1);
      expect(mockRequire).toHaveBeenCalledWith(
        [`${btModulesLoadConfig.hostedFields.amdModule.minified}`],
        expect.any(Function),
        expect.any(Function)
      );
    });

    describe("when ESM is NOT supported", () => {
      beforeAll(() => {
        mockIsEsmSupported.mockReturnValue(false);
      });

      it("should load the minified AMD modules", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockRequire).toHaveBeenCalledWith(
          [btModulesLoadConfig.hostedFields.amdModule.minified],
          expect.any(Function),
          expect.any(Function)
        );
        expect(mockLoadScriptWithExport).toHaveBeenCalled();
        expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
          expect.objectContaining({
            src: `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
          })
        );
      });

      it("should load the unminified AMD modules", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: false,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadScriptWithExport).toHaveBeenCalled();
        expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
          expect.objectContaining({
            src: `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.unminified}.js`,
          })
        );
      });

      it("should default to the minified script when no environment is specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadScriptWithExport).toHaveBeenCalled();
        expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
          expect.objectContaining({
            src: `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
          })
        );
      });

      it("should load the bundle override URL when specified", async () => {
        const bundleId = "2bc4343c92655";
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          metadata: { bundleIdOverride: bundleId },
        });

        await expect(promise).resolves.toStrictEqual({
          fastlane: mockFastlane,
          metadata: {
            localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
          },
        });
        expect(mockLoadScriptWithExport).toHaveBeenCalled();
        expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
          expect.objectContaining({
            src: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
          })
        );
      });

      it("should call loadScript with the minified URL and bundle override URL when specified", async () => {
        const bundleId = "9df9369c92655";
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
          metadata: { bundleIdOverride: bundleId },
        });

        await expect(promise).resolves.toStrictEqual({
          fastlane: mockFastlane,
          metadata: {
            localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
          },
        });
        expect(mockLoadScriptWithExport).toHaveBeenCalled();
        expect(mockLoadScriptWithExport).toHaveBeenCalledWith(
          expect.objectContaining({
            src: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ASSET_NAME.minified}.js`,
          })
        );
      });
    });

    describe("when ESM is supported", () => {
      beforeAll(() => {
        mockIsEsmSupported.mockReturnValue(true);
      });

      it("should load the minified AMD modules", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockRequire).toHaveBeenCalledWith(
          [btModulesLoadConfig.hostedFields.amdModule.minified],
          expect.any(Function),
          expect.any(Function)
        );
        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });

      it("should load the unminified AMD modules", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: false,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.unminified}.js`
        );
      });

      it("should default to the minified script when no environment is specified", async () => {
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
        });

        await expect(promise).resolves.toStrictEqual(
          testContext.mockLoadResponse
        );
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `${constants.CDNX_PROD}/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });

      it("should load the bundle override URL when specified", async () => {
        const bundleId = "2bc4343c92655";
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          metadata: { bundleIdOverride: bundleId },
        });

        await expect(promise).resolves.toStrictEqual({
          fastlane: mockFastlane,
          metadata: {
            localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
          },
        });
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });

      it("should call loadScript with the minified URL and bundle override URL when specified", async () => {
        const bundleId = "9df9369c92655";
        const promise = loadAxo({
          platform: AxoSupportedPlatforms.BT,
          btSdkVersion: version,
          minified: true,
          metadata: { bundleIdOverride: bundleId },
        });

        await expect(promise).resolves.toStrictEqual({
          fastlane: mockFastlane,
          metadata: {
            localeUrl: `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.LOCALE_PATH}`,
          },
        });
        expect(mockLoadModule).toHaveBeenCalled();
        expect(mockLoadModule).toHaveBeenCalledWith(
          `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${constants.AXO_ASSET_PATH}/${constants.AXO_ESM_PATH}/${constants.AXO_ASSET_NAME.minified}.js`
        );
      });
    });
  });
});
