import { constants, FL_NAMESPACE } from "./constants";
import {
  LoadAxoOptions,
  GenerateAssetUrlParams,
  AxoSupportedPlatforms,
} from "./types";
import { safeLoadBtModule } from "./lib/safeLoadBtModule";
import { btModulesLoadConfig } from "./lib/safeLoadBtModule/types";
import { isEsmSupported, loadModule } from "./lib/esm-utils";
import { loadScriptWithExport } from "./lib/script-utils";

/**
 * Loads accelerated checkout components.
 * @param options object with a minified parameter to determine if the script that is loaded should be minified or not (defaults to true if)
 * @returns an object with metadata with a localeUrl parameter to be read by AXO SDK
 */
async function loadAxo(options: LoadAxoOptions) {
  performance.mark("pp_axo_sdk_init_invoked");
  const { btSdkVersion, minified } = options;
  const assetUrl = getAssetPath(options);
  const localeUrl = getLocaleUrl(options);
  const enableAmdSupport = options.platform === AxoSupportedPlatforms.BT;

  if (options.platform === AxoSupportedPlatforms.BT) {
    await Promise.all([
      safeLoadBtModule(
        btModulesLoadConfig.hostedFields,
        btSdkVersion,
        enableAmdSupport,
        minified
      ),
    ]);
  } else if (options.platform === AxoSupportedPlatforms.PPCP) {
    await Promise.all([
      safeLoadBtModule(btModulesLoadConfig.client, btSdkVersion, minified),
      safeLoadBtModule(
        btModulesLoadConfig.hostedFields,
        btSdkVersion,
        minified
      ),
    ]);
  } else {
    throw new Error("unsupported axo platform");
  }

  const fastlane = await loadAXOScript(assetUrl, minified);

  return { fastlane, metadata: { localeUrl } };
}

/**
 * Reads the url and to load the axo bundle script
 * @param url (Required) string url for the correct axo asset
 * @returns Promise<HTMLScriptElement>
 */
async function loadAXOScript(url: string, minified = true) {
  if (isEsmSupported()) {
    const { default: fastlane } = await loadModule(url);
    window.braintree = window.braintree || {};
    window.braintree.fastlane = fastlane;
    return fastlane;
  }

  return loadScriptWithExport({
    id: "axo-id",
    src: url,
    name: "Fastlane",
    exportedName: "braintree.fastlane",
  });
}

/**
 * Prepends the domain to the asset url
 * @param options object with assetUrl and bundleid parameters to determine which URL to return
 * @returns full domain and assets URL as string
 */
function generateAssetUrl({
  assetPath,
  bundleId,
}: GenerateAssetUrlParams): string {
  return bundleId
    ? `https://cdn-${bundleId}.static.engineering.dev.paypalinc.com/${assetPath}`
    : `${constants.CDNX_PROD}/${assetPath}`;
}

/**
 * Retrieves either the minified or unminified assets URL as specified
 * @param options (Optional) object with a minified and metadata with bundleIdOverride parameters to determine which URL to return
 * @returns assets URL as string
 */
function getAssetPath(options?: LoadAxoOptions): string {
  const assetName =
    options?.minified !== false
      ? constants.AXO_ASSET_NAME.minified
      : constants.AXO_ASSET_NAME.unminified;

  const assetPath = `${constants.AXO_ASSET_PATH}/${
    isEsmSupported() ? constants.AXO_ESM_PATH + "/" : ""
  }${assetName}.js`;

  return generateAssetUrl({
    assetPath,
    bundleId: options?.metadata?.bundleIdOverride,
  });
}

/**
 * Retrieves the Locales URL, the path to our language files
 * @param options (Optional) object with a minified and metadata with bundleIdOverride parameters to determine which URL to return
 * @returns locale URL as string
 */
function getLocaleUrl(options?: LoadAxoOptions): string {
  return generateAssetUrl({
    assetPath: constants.LOCALE_PATH,
    bundleId: options?.metadata?.bundleIdOverride,
  });
}

export default loadAxo;
